import numpy as np

from asyncpg import _testbase as tb
from asyncpg.rkt import set_query_dtype


class TestRocket(tb.TestCase):

    def test_rkt_smoke(self):
        query = "SELECT 1, 1"
        dtype = np.dtype([("a", int), ("b", int)])
        augmented_query = set_query_dtype(query, dtype)
        self.assertEqual(augmented_query[:3], "--🚀")
        end = augmented_query.find("🚀\n", 1)
        header = augmented_query[3:end]
        self.assertEqual(augmented_query[end + 2:], query)
        for char in header:
            self.assertIn(char, "0123456789abcdef")

    def test_rkt_wrong_type(self):
        with self.assertRaises(TypeError):
            set_query_dtype("SELECT 1", 1)
        with self.assertRaises(ValueError):
            set_query_dtype("SELECT 1", None)

    def test_rkt_wrong_dtype(self):
        with self.assertRaises(ValueError):
            set_query_dtype("SELECT 1", np.int64)



from cpython.ref cimport PyObject
import cython
from libc.stdint cimport intptr_t, int16_t, int32_t, uint32_t, int64_t
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy, memset
from math import nan

import numpy as np
from numpy cimport PyArray_DATA


cdef extern from "utf8_to_ucs4.h":
    int utf8_to_ucs4(const char *, int32_t *, int) nogil


cdef extern from "memalign.h":
    void *memalign(size_t alignment, size_t size) nogil
    void aligned_free(void *) nogil


@cython.no_gc
@cython.final
@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.wraparound(False)
@cython.freelist(_BUFFER_FREELIST_SIZE)
cdef class ArrayWriter:
    def __cinit__(self, np_dtype dtype not None):
        cdef:
            np_dtype child_dtype
            int offset, pos = 0, length = len(dtype), unit
            libdivide_s64_ex_t adjust_value
            char adjust_kind
            object metadata

        if not dtype.fields:
            raise ValueError("dtype must be a struct")
        self.dtype = dtype
        metadata = dtype.metadata
        if cpythonx.PyMapping_Check(metadata) and cpythonx.PyMapping_HasKey(metadata, "blocks"):
            self.major = kColumnMajor
        else:
            self.major = kRowMajor
        self.null_indexes = []
        self._dtype_length = length
        self._chunks = []
        self._recharge()
        self._dtype_kind = <char *>malloc(length)
        self._dtype_size = <uint32_t *>malloc(length * 4)
        self._dtype_offset = <uint32_t *>malloc((length + (self.major == kRowMajor)) * 4)
        self._time_adjust_value = <libdivide_s64_ex_t *>malloc(length * sizeof(libdivide_s64_ex_t))
        for i, name in enumerate(dtype.names):
            child_dtype, offset = dtype.fields[name]
            self._dtype_kind[i] = child_dtype.kind
            self._dtype_size[i] = child_dtype.itemsize
            if self.major == kRowMajor:
                self._dtype_offset[i] = offset - pos
            elif i == 0:
                self._dtype_offset[i] = 0
            else:
                self._dtype_offset[i] = \
                    self._dtype_offset[i - 1] + _ARRAY_CHUNK_SIZE * self._dtype_size[i - 1]
            pos = offset + child_dtype.itemsize
            if child_dtype.kind == b"M" or child_dtype.kind == b"m":
                unit = (<PyArray_DatetimeDTypeMetaData *>child_dtype.c_metadata).meta.base
                if unit == NPY_FR_us or unit == NPY_FR_GENERIC:
                    adjust_value.base.magic = 1
                    adjust_value.dt_offset = 0
                elif unit == NPY_FR_ns:
                    adjust_value.base.magic = 1000
                    adjust_value.dt_offset = 0
                elif unit == NPY_FR_s:
                    adjust_value.base = libdivide_s64_gen(1000000)
                    adjust_value.dt_offset = 1 - 1000000
                elif unit == NPY_FR_ms:
                    adjust_value.base = libdivide_s64_gen(1000)
                    adjust_value.dt_offset = 1 - 1000
                elif unit == NPY_FR_D:
                    adjust_value.base = libdivide_s64_gen(24 * 3600 * 1000000)
                    adjust_value.dt_offset = 1 - 24 * 3600 * 1000000
                elif unit == NPY_FR_m:
                    adjust_value.base = libdivide_s64_gen(60 * 1000000)
                    adjust_value.dt_offset = 1 - 60 * 1000000
                elif unit == NPY_FR_h:
                    adjust_value.base = libdivide_s64_gen(3600 * 1000000)
                    adjust_value.dt_offset = 1 - 3600 * 1000000
                elif unit == NPY_FR_W:
                    adjust_value.base = libdivide_s64_gen(7 * 24 * 3600 * 1000000)
                    adjust_value.dt_offset = 1 - 7 * 24 * 3600 * 1000000
                else:
                    raise NotImplementedError(f"dtype[{i}] = {dtype[i]} time unit is not supported")
            self._time_adjust_value[i] = adjust_value
        if self.major == kRowMajor:
            self._dtype_offset[length] = dtype.itemsize - pos

    def __dealloc__(self):
        free(self._dtype_kind)
        free(self._dtype_size)
        free(self._dtype_offset)
        free(self._time_adjust_value)
        for chunk in self._chunks:
            aligned_free(cpythonunsafe.PyLong_AsVoidPtr(<PyObject *>chunk))

    cdef void _step(self):
        if self.major == kRowMajor:
            self._data += self._dtype_size[self._field]
        self._field += 1
        if self._field < self._dtype_length:
            if self.major == kRowMajor:
                self._data += self._dtype_offset[self._field]
            else:
                self._data = (
                    self._chunk
                    + self._dtype_offset[self._field]
                    + self._item * self._dtype_size[self._field]
                )
        if self._field == self._dtype_length:
            self._field = 0
            self._item += 1
            if self.major == kColumnMajor:
                self._data = self._chunk + self._item * self._dtype_size[self._field]
            if self._item == _ARRAY_CHUNK_SIZE:
                self._recharge()

    cdef void _recharge(self):
        self._item = 0
        self._data = self._chunk = <char *> memalign(4096, _ARRAY_CHUNK_SIZE * self.dtype.itemsize)
        self._chunks.append(<intptr_t> self._chunk)

    cdef consolidate(self):
        if self.major == kRowMajor:
            result = self._consolidate_row_major()
        else:
            result = self._consolidate_column_major()
        for chunk in self._chunks:
            aligned_free(cpythonunsafe.PyLong_AsVoidPtr(<PyObject *>chunk))
        self._chunks.clear()
        return result

    cdef _consolidate_row_major(self):
        cdef:
            char *body
            void *chunk
            np_dtype dtype = self.dtype
            int chunks_count = len(self._chunks), i
            int64_t chunk_size = _ARRAY_CHUNK_SIZE * dtype.itemsize

        arr = np.empty((chunks_count - 1) * _ARRAY_CHUNK_SIZE + self._item, dtype=self.dtype)
        body = <char*> PyArray_DATA(arr)
        for i in range(chunks_count):
            chunk = cpythonunsafe.PyLong_AsVoidPtr(
                cpythonunsafe.PyList_GET_ITEM(<PyObject*>self._chunks, i))
            if i == chunks_count - 1:
                chunk_size = self._item * dtype.itemsize
            memcpy(body, chunk, chunk_size)
            body += chunk_size
        return arr

    cdef _consolidate_column_major(self):
        cdef:
            char *body
            char *chunk
            np_dtype dtype = self.dtype, child_type
            int last_chunk_index = len(self._chunks) - 1, i, j
            int64_t copy_size
            int64_t total_items = last_chunk_index * _ARRAY_CHUNK_SIZE + self._item
            dict blocks = {}
            list indexes
        arr = np.empty(1, dtype=[(name, object) for name in self.dtype.names])[0]
        # group same child dtypes into blocks
        for i in range(self._dtype_length):
            indexes = blocks.setdefault(self.dtype[i], [])
            indexes.append(i)
        for child_dtype, indexes in blocks.items():
            block = np.empty((len(indexes), total_items), dtype=child_dtype)
            for i, j in enumerate(indexes):
                arr[j] = block[i]
        # memcpy chunk by chunk, column by column
        for j in range(last_chunk_index + 1):
            chunk = <char *> cpythonunsafe.PyLong_AsVoidPtr(
                cpythonunsafe.PyList_GET_ITEM(<PyObject *> self._chunks, j))
            for i in range(self._dtype_length):
                copy_size = _ARRAY_CHUNK_SIZE * self._dtype_size[i]
                body = <char*> PyArray_DATA(arr[i]) + j * copy_size
                if j == last_chunk_index:
                    copy_size = self._item * self._dtype_size[i]
                memcpy(body, chunk + self._dtype_offset[i], copy_size)
        return arr

    cdef raise_dtype_error(self, str pgtype, int size):
        raise DTypeError(f"dtype[{self._field}] = {self.dtype[self._field]} does not match "
                         f"PostgreSQL {pgtype}" + (f" of size {size}" if size > 0 else ""))

    cdef inline int current_field_is_object(self) nogil:
        return self._dtype_kind[self._field] == b"O"

    cdef int write_null(self) except -1:
        cdef:
            int i
            char dtype
            int size

        self.null_indexes.append(
            (_ARRAY_CHUNK_SIZE * (len(self._chunks) - 1) + self._item) * self._dtype_length
            + self._field)
        dtype = self._dtype_kind[self._field]
        size = self._dtype_size[self._field]
        if dtype == b"O":
            (<PyObject **> self._data)[0] = cpythonunsafe.Py_None
            cpythonunsafe.Py_INCREF(cpythonunsafe.Py_None)
        elif dtype == b"M" or dtype == b"m":
            (<int64_t *>self._data)[0] = (1 << 63)  # NaT
        elif dtype == b"f":
            if size == 4:
                (<float *> self._data)[0] = <float>nan
            elif size == 8:
                (<double *> self._data)[0] = nan
        elif dtype == b"V":
            if size % 8 == 0:
                for i in range(size // 8):
                    (<double *> self._data)[i] = nan
            else:
                for i in range(size):
                    (<uint8_t *> self._data)[i] = (1 << 7)
        elif dtype == b"i" or dtype == b"u" or dtype == b"b":
            if size == 1:
                (<uint8_t *> self._data)[0] = (1 << 7)
            elif size == 2:
                (<uint16_t *> self._data)[0] = (1 << 15)
            elif size == 4:
                (<uint32_t *> self._data)[0] = (1 << 31)
            elif size == 8:
                (<uint64_t *> self._data)[0] = (1 << 63)
        elif dtype == b"S":
            memset(self._data, 0xFF, size)
        elif dtype == b"U":
            memset(self._data, 0, size)
        self._step()

    cdef int write_object(self, object obj) except -1:
        cdef:
            PyObject *ptr = <PyObject *>obj
            int ret
        ret = self.write_object_unsafe(ptr)
        cpythonunsafe.Py_INCREF(ptr)
        return ret

    cdef int write_object_unsafe(self, PyObject *obj) except -1:
        if not self.current_field_is_object():
            self.raise_dtype_error("object", 0)
        (<PyObject **> self._data)[0] = obj
        self._step()

    cdef int write_bool(self, int b) except -1:
        if self._dtype_kind[self._field] != b"b":
            self.raise_dtype_error("bool", 1)
        self._data[0] = b != 0
        self._step()

    cdef int write_bytes(self, const char *data, ssize_t len) except -1:
        cdef int full_size = self._dtype_size[self._field]
        if full_size < len or self._dtype_kind[self._field] != b"S":
            self.raise_dtype_error("bytea", len)
        memcpy(self._data, data, len)
        memset(self._data + len, 0, full_size - len)
        self._step()

    cdef int write_string(self, const char *data, ssize_t len) except -1:
        cdef:
            char kind = self._dtype_kind[self._field]
            int full_size = self._dtype_size[self._field]
            int ucs4_size
        if kind == b"U":
            if full_size < 4 * len:
                self.raise_dtype_error("text", len)
            ucs4_size = 4 * utf8_to_ucs4(data, <int32_t *> self._data, len)
            memset(self._data + ucs4_size, 0, full_size - ucs4_size)
        elif kind == b"S":
            if full_size < len:
                self.raise_dtype_error("text", len)
            memcpy(self._data, data, len)
            memset(self._data + len, 0, full_size - len)
        else:
            self.raise_dtype_error("text", len)
        self._step()

    cdef int write_int16(self, int16_t i) except -1:
        cdef char kind = self._dtype_kind[self._field]
        if (kind != b"i" and kind != b"u") or self._dtype_size[self._field] != 2:
            self.raise_dtype_error("smallint", 2)
        (<int16_t *> self._data)[0] = i
        self._step()

    cdef int write_int32(self, int32_t i) except -1:
        cdef char kind = self._dtype_kind[self._field]
        if (kind != b"i" and kind != b"u") or self._dtype_size[self._field] != 4:
            self.raise_dtype_error("int", 4)
        (<int32_t *> self._data)[0] = i
        self._step()

    cdef int write_int64(self, int64_t i) except -1:
        cdef char kind = self._dtype_kind[self._field]
        if (kind != b"i" and kind != b"u") or self._dtype_size[self._field] != 8:
            self.raise_dtype_error("bigint", 8)
        (<int64_t *> self._data)[0] = i
        self._step()

    cdef int write_float(self, float f) except -1:
        if self._dtype_kind[self._field] != b"f" or self._dtype_size[self._field] != 4:
            self.raise_dtype_error("float4", 4)
        (<float *> self._data)[0] = f
        self._step()

    cdef int write_double(self, double d) except -1:
        if self._dtype_kind[self._field] != b"f" or self._dtype_size[self._field] != 8:
            self.raise_dtype_error("float8", 8)
        (<double *> self._data)[0] = d
        self._step()

    cdef int write_datetime(self, int64_t dt) except -1:
        cdef:
            libdivide_s64_ex_t *ptr = &self._time_adjust_value[self._field]
            int64_t offset = ptr.dt_offset
        if self._dtype_kind[self._field] != b"M":
            self.raise_dtype_error("timestamp", 8)
        if offset != 0:
            if dt < 0:
                dt += offset
            dt = libdivide_s64_do(dt, <libdivide_s64_t *>ptr)
        else:
            dt *= ptr.base.magic
        (<int64_t *> self._data)[0] = dt
        self._step()

    cdef int write_timedelta(self, int64_t td) except -1:
        cdef libdivide_s64_ex_t *ptr = &self._time_adjust_value[self._field]
        if self._dtype_kind[self._field] != b"m":
            self.raise_dtype_error("time", 8)
        if ptr.dt_offset != 0:
            td = libdivide_s64_do(td, <libdivide_s64_t *>ptr)
        else:
            td *= ptr.base.magic
        (<int64_t *> self._data)[0] = td
        self._step()

    cdef int write_4d(self, double high_x, double high_y, double low_x, double low_y) except -1:
        if self._dtype_kind[self._field] != b"V" or self._dtype_size[self._field] != 8 * 4:
            self.raise_dtype_error("4 of float8", 32)
        (<double *> self._data)[0] = high_x
        (<double *> self._data)[1] = high_y
        (<double *> self._data)[2] = low_x
        (<double *> self._data)[3] = low_y
        self._step()

    cdef int write_3d(self, double a, double b, double c) except -1:
        if self._dtype_kind[self._field] != b"V" or self._dtype_size[self._field] != 8 * 3:
            self.raise_dtype_error("3 of float8", 24)
        (<double *> self._data)[0] = a
        (<double *> self._data)[1] = b
        (<double *> self._data)[2] = c
        self._step()

    cdef int write_2d(self, double x, double y) except -1:
        if self._dtype_kind[self._field] != b"V" or self._dtype_size[self._field] != 8 * 2:
            self.raise_dtype_error("2 of float8", 16)
        (<double *> self._data)[0] = x
        (<double *> self._data)[1] = y
        self._step()

    cdef int write_tid(self, uint32_t block, uint16_t offset) except -1:
        if self._dtype_kind[self._field] != b"V" or self._dtype_size[self._field] != (4 + 2):
            self.raise_dtype_error("tid", 6)
        (<uint32_t *> self._data)[0] = block
        (<uint16_t *> self._data)[2] = offset
        self._step()



import py

from .compat import Iterable
from .utils import TIME_UNITS
from .utils import slugify

try:
    from pygal.graph.box import Box
    from pygal.style import DefaultStyle
except ImportError as exc:
    raise ImportError(exc.args, "Please install pygal and pygaljs or pytest-benchmark[histogram]")


class CustomBox(Box):
    def _box_points(self, serie, _):
        return serie, [serie[0], serie[6]]

    def _value_format(self, x):
        return "Min: {0[8]:.4f}\n" \
               "Q1-1.5IQR: {0[1]:.4f}\n" \
               "Q1: {0[2]:.4f}\nMedian: {0[3]:.4f}\nQ3: {0[4]:.4f}\n" \
               "Q3+1.5IQR: {0[5]:.4f}\n" \
               "Max: {0[9]:.4f}".format(x[:10])

    def _format(self, x, *args):
        sup = super(CustomBox, self)._format
        if args:
            val = x.values
        else:
            val = x
        if isinstance(val, Iterable):
            return self._value_format(val), val[7]
        else:
            return sup(x, *args)

    def _tooltip_data(self, node, value, x, y, classes=None, xlabel=None):
        super(CustomBox, self)._tooltip_data(node, value[0], x, y, classes=classes, xlabel=None)
        self.svg.node(node, 'desc', class_="x_label").text = value[1]


def make_plot(benchmarks, title, adjustment):
    class Style(DefaultStyle):
        guide_stroke_color = "rgba(0,0,0,0.4)"
        # colors = ["#000000" if row["path"] else DefaultStyle.colors[1]
        #          for row in benchmarks]
        colors = []
        for row in benchmarks:
            i = 1
            name = row["name"].replace("test_", "")
            if name == "min":
                i = 16
            elif name.startswith("record"):
                i = 4
            elif name.startswith("numpy"):
                i = 7
            if row["path"]:
                color = "#000000"
            else:
                color = DefaultStyle.colors[i]
            colors.append(color)

        font_family = 'Consolas, "Deja Vu Sans Mono", "Bitstream Vera Sans Mono", "Courier New", monospace'

    minimum = int(min(row["min"] * adjustment for row in benchmarks))
    maximum = int(max(
        min(row["max"], row["hd15iqr"]) * adjustment
        for row in benchmarks
    ) + 1)

    try:
        import pygaljs
    except ImportError:
        opts = {}
    else:
        opts = {
            "js": [
                pygaljs.uri("2.0.x", "pygal-tooltips.js")
            ]
        }
    x_labels = []
    for row in benchmarks:
        name = row["name"].replace("test_", "")
        if (param_pos := name.find("[")) >= 0:
            param = name[param_pos + 1:-1]
            name = name[:param_pos]
            label=f"{name}%{param}"
        else:
            label = name
        # <text x="([^"]+)"([^>]+)>([^<%]+)%([^<]+)
        # <text x="$1"$2><tspan x="$1" alignment-baseline="central">$3</tspan><tspan x="$1" alignment-baseline="central" dy="12">$4</tspan>
        x_labels.append(label)

    plot = CustomBox(
        box_mode="tukey",
        # x_label_rotation=-90,
        x_labels=x_labels,
        show_legend=False,
        title=title,
        x_title="Trial",
        y_title="Duration",
        style=Style,
        min_scale=20,
        max_scale=20,
        truncate_label=50,
        range=(0, maximum),
        zero=minimum,
        css=[
            "file://style.css",
            "file://graph.css",
            """inline:
                .tooltip .value {
                    font-size: 1em !important;
                }
                .axis text {
                    font-size: 9px !important;
                }
            """
        ],
        **opts
    )

    for row in benchmarks:
        serie = []
        for field in ["min", "ld15iqr", "q1", "median", "q3", "hd15iqr", "max"]:
            if field not in ("min", "max"):
                value = row[field] * adjustment
            else:
                value = -maximum / 40
            serie.append(value)
        serie.append(row["path"])
        for ex in ("min", "max"):
            serie.append(row[ex] * adjustment)
        plot.add("{0[fullname]} - {0[rounds]} rounds".format(row), serie)
    return plot


def make_histogram(output_prefix, name, benchmarks, unit, adjustment):
    if name:
        path = "{0}-{1}.svg".format(output_prefix, slugify(name))
        title = "Speed in {0} of {1}".format(TIME_UNITS[unit], name)
    else:
        path = "{0}.svg".format(output_prefix)
        title = "Speed in {0}".format(TIME_UNITS[unit])

    output_file = py.path.local(path).ensure()

    plot = make_plot(
        benchmarks=benchmarks,
        title=title,
        adjustment=adjustment,
    )
    plot.render_to_file(str(output_file))
    return output_file



import pickle
from typing import Union

import numpy as np


def set_query_dtype(query: str, dtype: Union[bytes, str, np.dtype]) -> str:
    """
    Augment the query string with the result numpy data type.

    :param query: SQL query to augment.
    :param dtype: numpy dtype-compatible or pickled numpy dtype. \
                  The dtype must be structured.
    :return: Augmented SQL query with the embedded dtype.
    """
    if isinstance(dtype, bytes):
        # assume already serialized
        serialized_dtype = dtype.hex()
    else:
        dtype = np.dtype(dtype)
        if not dtype.fields:
            raise ValueError("The data type must be a structure")
        # we cannot write a pointer because the object is not referenced
        # directly from the string and may die
        serialized_dtype = pickle.dumps(dtype).hex()
    return f"--🚀{serialized_dtype}🚀\n{query}"



from cpython cimport PyObject
from libc.stdint cimport uint8_t, int64_t


cdef enum NPY_DATETIMEUNIT:
    # Force signed enum type, must be -1 for code compatibility
    NPY_FR_ERROR = -1,  # error or undetermined

    # Start of valid units
    NPY_FR_Y = 0,  # Years
    NPY_FR_M = 1,  # Months
    NPY_FR_W = 2,  # Weeks
    # Gap where 1.6 NPY_FR_B (value 3) was
    NPY_FR_D = 4,  # Days
    NPY_FR_h = 5,  # hours
    NPY_FR_m = 6,  # minutes
    NPY_FR_s = 7,  # seconds
    NPY_FR_ms = 8,  # milliseconds
    NPY_FR_us = 9,  # microseconds
    NPY_FR_ns = 10,  # nanoseconds
    NPY_FR_ps = 11,  # picoseconds
    NPY_FR_fs = 12,  # femtoseconds
    NPY_FR_as = 13,  # attoseconds
    NPY_FR_GENERIC = 14  # unbound units, can convert to anything


cdef extern from "numpy/arrayobject.h":
    ctypedef struct PyArray_DatetimeMetaData:
        NPY_DATETIMEUNIT base

    ctypedef struct PyArray_DatetimeDTypeMetaData:
        PyArray_DatetimeMetaData meta

    ctypedef class numpy.dtype[object PyArray_Descr, check_size ignore]:
        cdef:
            char kind
            int itemsize "elsize"
            dict fields
            tuple names
            void *c_metadata


ctypedef dtype np_dtype

cdef enum StorageMajor:
    kRowMajor
    kColumnMajor


cdef extern from "numpy/libdivide/libdivide.h":
    struct libdivide_s64_t:
        int64_t magic
        uint8_t more

    libdivide_s64_t libdivide_s64_gen(int64_t d) nogil
    int64_t libdivide_s64_do(int64_t numer, const libdivide_s64_t *denom) nogil


cdef class DTypeError(Exception):
    pass


cdef struct libdivide_s64_ex_t:
    libdivide_s64_t base
    int64_t dt_offset


cdef class ArrayWriter:
    cdef:
        np_dtype dtype
        StorageMajor major
        list null_indexes
        list _chunks
        int _dtype_length
        char *_dtype_kind
        uint32_t *_dtype_size
        uint32_t *_dtype_offset
        libdivide_s64_ex_t *_time_adjust_value
        int64_t _item
        int16_t _field
        char *_data
        char *_chunk

    cdef raise_dtype_error(self, str pgtype, int size)
    cdef int current_field_is_object(self) nogil
    cdef int write_null(self) except -1
    cdef int write_object(self, object obj) except -1
    cdef int write_object_unsafe(self, PyObject *obj) except -1
    cdef int write_bool(self, int b) except -1
    cdef int write_bytes(self, const char *data, ssize_t len) except -1
    cdef int write_string(self, const char *data, ssize_t len) except -1
    cdef int write_int16(self, int16_t i) except -1
    cdef int write_int32(self, int32_t i) except -1
    cdef int write_int64(self, int64_t i) except -1
    cdef int write_float(self, float f) except -1
    cdef int write_double(self, double d) except -1
    cdef int write_datetime(self, int64_t dt) except -1
    cdef int write_timedelta(self, int64_t td) except -1
    cdef int write_4d(self, double high_x, double high_y, double low_x, double low_y) except -1
    cdef int write_3d(self, double a, double b, double c) except -1
    cdef int write_2d(self, double x, double y) except -1
    cdef int write_tid(self, uint32_t block, uint16_t offset) except -1

    cdef consolidate(self)
    cdef _consolidate_row_major(self)
    cdef _consolidate_column_major(self)
    cdef void _step(self)
    cdef void _recharge(self)



from cpython cimport PyObject

cdef extern from "Python.h":
    PyObject *PyBytes_FromStringAndSize(const char *, ssize_t)
    PyObject *PyUnicode_DecodeUTF8(const char *, Py_ssize_t, const char *)
    void Py_INCREF(PyObject *o)
    PyObject *PyTuple_GET_ITEM(PyObject *, Py_ssize_t) nogil
    PyObject *PyList_GET_ITEM(PyObject *, Py_ssize_t) nogil
    void *PyLong_AsVoidPtr(PyObject *pylong)
    PyObject *PyLong_FromLong(long v)
    PyObject *PyLong_FromLongLong(long long v)
    PyObject *Py_None
    PyObject *Py_True
    PyObject *Py_False



from datetime import date, datetime, time, timedelta
import decimal
import ipaddress
import random
from uuid import UUID

import numpy as np
from numpy.testing import assert_array_equal

from asyncpg import _testbase as tb, DTypeError
from asyncpg.rkt import set_query_dtype


type_samples = [
    ("bool_false", False, "false", np.bool8),
    ("bool_true", True, "true", np.bool8),
    ("bool_obj", True, "true", object),
    ("bool_null", True, "null", np.bool8),

    ("int2_32000", 32000, "32000::smallint", np.int16),
    ("int2_1024", -1024, "-1024::smallint", np.int16),
    ("int2_0", 0, "0::smallint", np.int16),
    ("int2_obj", 32000, "32000::smallint", object),
    ("int2_null", -(1 << 15), "null", np.int16),

    ("uint2_32000", 32000, "32000::smallint", np.uint16),
    ("uint2_1024", 64512, "-1024::smallint", np.uint16),
    ("uint2_0", 0, "0::smallint", np.uint16),
    ("uint2_null", 1 << 15, "null", np.uint16),

    ("int4_320000", 3200000, "3200000::int", np.int32),
    ("int4_100024", -100024, "-100024::int", np.int32),
    ("int4_0", 0, "0::int", np.int32),
    ("int4_obj", 3200000, "3200000::int", object),
    ("int4_null", -(1 << 31), "null", np.int32),

    ("uint4_320000", 3200000, "3200000::int", np.uint32),
    ("uint4_100024", 4294867272, "-100024::int", np.uint32),
    ("uint4_0", 0, "0::int", np.uint32),
    ("uint4_null", 1 << 31, "null", np.uint32),

    ("int8_32000000000", 32000000000, "32000000000::bigint", np.int64),
    ("int8_10000000024", -10000000024, "-10000000024::bigint", np.int64),
    ("int8_0", 0, "0::bigint", np.int64),
    ("int8_obj", 32000000000, "32000000000::bigint", object),
    ("int8_null", -(1 << 63), "null", np.int64),

    ("uint8_32000000000", 32000000000, "32000000000::bigint", np.uint64),
    ("uint8_10000000024", 18446744063709551592, "-10000000024::bigint",
     np.uint64),
    ("uint8_0", 0, "0::bigint", np.uint64),
    ("uint8_null", 1 << 63, "null", np.uint64),

    ("float", 2.125, "2.125::float4", np.float32),
    ("floatnan", np.nan, "'NaN'::float4", np.float32),
    ("floatinf", np.inf, "'inf'::float4", np.float32),
    ("floatneginf", -np.inf, "-'inf'::float4", np.float32),
    ("floatnull", np.nan, "null", np.float32),
    ("floatobj", 2.125, "2.125::float4", object),

    ("double", 2.125, "2.125::float8", np.float64),
    ("doublenan", np.nan, "'NaN'::float8", np.float64),
    ("doubleinf", np.inf, "'inf'::float8", np.float64),
    ("doubleneginf", -np.inf, "-'inf'::float8", np.float64),
    ("doublenull", np.nan, "null", np.float64),
    ("doubleobj", 2.125, "2.125::float8", object),

    ("numeric", decimal.Decimal(10000), "10000::numeric", object),

    ("bytea0", b"", "''::bytea", "S1"),
    ("bytea5", b"12345", "'12345'::bytea", "S5"),
    ("bytea10", b"12345", "'12345'::bytea", "S10"),
    ("byteanull", b"\xff", "null", "S1"),
    ("byteaobj", b"12345", "'12345'::bytea", object),

    ("text0", "", "''::text", "U1"),
    ("text5", "12345", "'12345'::text", "U5"),
    ("text10", "12345", "'12345'::text", "U10"),
    ("textrocket", "🚀áño", "'🚀áño'::text", "U9"),
    ("textnull", "", "null", "U1"),
    ("textobj", "12345", "'12345'::text", object),

    ("char", b"1", "'1'::char", "S1"),

    ("dtslt", np.datetime64("1989-01-12 12:00:01", "s"),
     "'1989-01-12 12:00:01.123'::timestamp", "datetime64[s]"),
    ("dtsgt", np.datetime64("1989-01-12 12:00:01", "s"),
     "'1989-01-12 12:00:01.723'::timestamp", "datetime64[s]"),
    ("dtms", np.datetime64("1989-01-12 12:00:01.123", "ms"),
     "'1989-01-12 12:00:01.123'::timestamp", "datetime64[ms]"),
    ("dtsz", np.datetime64("1989-01-12 11:00:01", "s"),
     "'1989-01-12 12:00:01.123+01:00'::timestamptz", "datetime64[s]"),
    ("dtmsz", np.datetime64("1989-01-12 11:00:01.123", "ms"),
     "'1989-01-12 12:00:01.123+01:00'::timestamptz", "datetime64[ms]"),
    ("dt1800lt", np.datetime64("1800-01-12 12:00:01", "s"),
     "'1800-01-12 12:00:01.123'::timestamp", "datetime64[s]"),
    ("dt1800gt", np.datetime64("1800-01-12 12:00:01", "s"),
     "'1800-01-12 12:00:01.723'::timestamp", "datetime64[s]"),
    ("dt1800ex", np.datetime64("1800-01-12 12:00:01", "s"),
     "'1800-01-12 12:00:01'::timestamp", "datetime64[s]"),
    ("dtmax", np.datetime64("294247-01-10T04:00:54.775807", "us"),
     "'infinity'::timestamp", "datetime64[us]"),
    ("dtmin", np.datetime64("-290308-12-21T19:59:05.224193", "us"),
     "'-infinity'::timestamp", "datetime64[us]"),
    ("dtnull", np.datetime64("NaT", "us"), "null", "datetime64[us]"),
    ("dtobj", datetime(1989, 1, 12, 12, 0, 1, microsecond=123000),
     "'1989-01-12 12:00:01.123'::timestamp", object),

    ("date", np.datetime64("1989-01-12", "D"),
     "'1989-01-12'::date", "datetime64[D]"),
    ("datenull", np.datetime64("NaT", "D"), "null", "datetime64[D]"),
    ("dateobj", date(1989, 1, 12), "'1989-01-12'::date", object),

    ("time100", np.timedelta64(100, "s"), "'00:01:40.123'::time",
     "timedelta64[s]"),
    ("time_100", np.timedelta64(-100, "s"), "-'00:01:40.123'::time",
     "timedelta64[s]"),
    ("time0", np.timedelta64(0, "us"), "'00:00:00'::time",
     "timedelta64[us]"),
    ("timenull", np.timedelta64("NaT", "us"), "null", "timedelta64[us]"),
    ("timeobj", time(0, 1, 40, 123000), "'00:01:40.123'::time",
     object),

    ("timetz100", np.timedelta64(-3500, "s"), "'00:01:40+01:00'::timetz",
     "timedelta64[s]"),

    ("interval", np.timedelta64(10, "D"), "'10 days'::interval",
     "timedelta64[D]"),
    ("intervalobj", timedelta(days=10), "'10 days'::interval", object),

    ("uuid", b"\x07" * 16, "'07070707-0707-0707-0707-070707070707'::uuid",
     "S16"),
    ("uuidobj", UUID(bytes=b"\x07" * 16),
     "'07070707-0707-0707-0707-070707070707'::uuid", object),

    ("varbit8", b"\x10", "'00010000'::varbit", "S1"),
    ("varbit9", b"\x10\x80", "'000100001'::varbit", "S2"),
    ("varbitobj", b"\x10", "'00010000'::varbit", object),

    ("tid", (10, 20), "'(10, 20)'::tid",
     np.dtype([("major", np.int32), ("minor", np.int16)])),

    ("oid", 987123, "987123::oid", np.uint32),

    ("point", (1.0, 2.0), "'(1.0, 2.0)'::point",
     np.dtype([("x", float), ("y", float)])),
    ("box", (3.0, 4.0, 1.0, 2.0), "'(1.0, 2.0, 3.0, 4.0)'::box",
     np.dtype([("high_x", float), ("high_y", float),
               ("low_x", float), ("low_y", float)])),
    ("lseg", (1.0, 2.0, 3.0, 4.0), "'(1.0, 2.0, 3.0, 4.0)'::lseg",
     np.dtype([("high_x", float), ("high_y", float),
               ("low_x", float), ("low_y", float)])),
    ("line", (1.0, 2.0, 3.0), "'{1.0, 2.0, 3.0}'::line",
     np.dtype([("a", float), ("b", float), ("c", float)])),
    ("circle", (1.0, 2.0, 3.0), "'1.0, 2.0, 3.0'::circle",
     np.dtype([("a", float), ("b", float), ("c", float)])),

    ("inet", ipaddress.IPv4Address('127.0.0.1'), "'127.0.0.1'::inet",
     object),
]

error_type_samples = [
    ("sanity", "1::int", np.int32, None),
    ("int", "7::int", bool, DTypeError),
    ("int64", "7::bigint", np.int32, DTypeError),
    ("int64", "7::bigint", "datetime64[s]", DTypeError),
    ("bytea", "'1234'::bytea", "S3", DTypeError),
    ("text", "'1234'::text", "U3", DTypeError),
    ("float", "1.0::float8", np.float32, DTypeError),
    ("varbit9", "'000100001'::varbit", "S1", DTypeError),
    ("dts", "'1989-01-12 12:00:01.123'::timestamp", int, DTypeError),
]


class TestCodecsNumpy(tb.ConnectedTestCase):

    def setup_standard_codecs(self, blocks: bool):
        dtype_body = []
        value_strs = []
        baseline = []
        nulls = []
        nans = []
        shuffled_type_samples = type_samples.copy()
        random.shuffle(shuffled_type_samples)
        for i, (name, value, value_sql, value_dtype) in enumerate(
                shuffled_type_samples):
            baseline.append(value)
            value_strs.append(value_sql)
            dtype_body.append((name, value_dtype))
            if value_sql == "null":
                nulls.append(i)
            if value != value:
                nans.append(i)
        baseline = tuple(baseline)
        if blocks:
            dtype = np.dtype(dtype_body, metadata={"blocks": True})
        else:
            dtype = np.dtype(dtype_body)
        return dtype, baseline, value_strs, nulls, nans

    async def test_standard_codecs_row(self):
        """Test decoding of standard data types to numpy arrays, row-major."""
        dtype, baseline, value_strs, nulls, nans = \
            self.setup_standard_codecs(False)
        for length in (0, 1, 512, 513, 1024):
            query = f"SELECT {', '.join(value_strs)}\n" \
                    f"FROM generate_series(1, {length}) i"

            stmt = await self.con.prepare(
                set_query_dtype(query, dtype)
            )
            with self.subTest(length=length):
                fetched_array, fetched_nulls = await stmt.fetch()

                self.assertIsInstance(fetched_nulls, list)
                self.assertEqual(len(fetched_nulls), len(nulls) * length)
                if length > 0:
                    self.assertEqual(fetched_nulls[:len(nulls)], nulls)
                self.assertIsInstance(fetched_array, np.ndarray)
                nan_mask = np.zeros(len(dtype), dtype=bool)
                nan_mask[nulls] = True
                nan_mask[nans] = True
                baseline_array = np.array([baseline] * length, dtype=dtype)
                # https://github.com/numpy/numpy/issues/21539
                for is_nan, (key, (child_dtype, _)) in zip(
                        nan_mask, dtype.fields.items()):
                    if not is_nan:
                        continue
                    if child_dtype.kind in ("m", "M", "f"):
                        assert_array_equal(
                            fetched_array[key], baseline_array[key])
                        fetched_array[key] = baseline_array[key] = 0
                assert_array_equal(fetched_array, baseline_array)

    async def test_standard_codecs_column(self):
        """Test decoding of standard data types to numpy arrays,
        column-major."""
        dtype, baseline, value_strs, nulls, _ = \
            self.setup_standard_codecs(True)
        for length in (0, 1, 512, 513, 1024):
            query = f"SELECT {', '.join(value_strs)}\n" \
                    f"FROM generate_series(1, {length}) i"

            stmt = await self.con.prepare(
                f"--skip\n{set_query_dtype(query, dtype)}"
            )
            with self.subTest(length=length):
                fetched_array, fetched_nulls = await stmt.fetch()

                self.assertIsInstance(fetched_nulls, list)
                self.assertEqual(len(fetched_nulls), len(nulls) * length)
                if length > 0:
                    self.assertEqual(fetched_nulls[:len(nulls)], nulls)
                    if length > 1:
                        fetched_nulls_chunk = fetched_nulls[
                            len(nulls):len(nulls) * 2
                        ]
                        nulls_chunk = (
                            np.array(nulls) + len(dtype)
                        ).tolist()
                        self.assertEqual(
                            fetched_nulls_chunk,
                            nulls_chunk)
                        if length > 512:
                            fetched_nulls_chunk = fetched_nulls[
                                len(nulls) * 512:len(nulls) * 513
                            ]
                            nulls_chunk = (
                                np.array(nulls) + 512 * len(dtype)
                            ).tolist()
                            self.assertEqual(
                                fetched_nulls_chunk,
                                nulls_chunk)
                self.assertIsInstance(fetched_array, np.void)
                for i, name in enumerate(dtype.names):
                    baseline_array = np.array(
                        [baseline[i]] * length, dtype=dtype.fields[name][0])
                    assert_array_equal(fetched_array[i], baseline_array)

    async def test_exceptions(self):
        """Test decoding of standard data types to numpy arrays."""
        for name, value_sql, value_dtype, exc in error_type_samples:
            dtype = np.dtype([(name, value_dtype)])
            query = set_query_dtype(f"SELECT {value_sql}", dtype)
            with self.subTest(name=name):
                if exc is not None:
                    with self.assertRaises(exc):
                        await self.con.fetch(query)
                else:
                    await self.con.fetch(query)

    async def test_empty_query(self):
        """Test return nothing with an empty query and a dtype."""
        dtype = np.dtype([("sha", "S40")])
        result = await self.con.fetch(set_query_dtype("", dtype))
        assert_array_equal(result[0], np.array([], dtype=dtype))
        assert result[1] == []



import pickle
import unittest

import numpy as np
try:
    import pytest
except ImportError:
    raise unittest.SkipTest("This is a benchmark")

from asyncpg.rkt import set_query_dtype
from asyncpg._testbase import ConnectedTestCase


def setup_suite():
    ConnectedTestCase.setUpClass()
    suite = ConnectedTestCase()
    suite._testMethodName = "setUp"
    suite.setUp()
    suite.stmt = None
    return suite


def _entry(suite, coro, length):
    suite.loop.run_until_complete(coro(suite, length))


async def _round_trip(suite, _):
    if suite.stmt is None:
        suite.stmt = await suite.con.prepare("SELECT 1")
    await suite.stmt.fetch()


@pytest.mark.benchmark(
    warmup=True,
    warmup_iterations=2,
)
@pytest.mark.parametrize("length", [1])
def test_min(benchmark, length):
    suite = setup_suite()
    benchmark(_entry, suite, _round_trip, length)


type_samples = [
    ("bool1", "false", np.bool8),
    ("bool2", "true", np.bool8),

    ("int1", "32000000000::bigint", np.int64),
    ("int2", "32000000001::bigint", np.int64),
    ("int3", "32000000002::bigint", np.int64),
    ("int4", "32000000003::bigint", np.int64),
    ("int5", "32000000004::bigint", np.int64),

    ("float", "2.125::float4", np.float32),

    ("bytes1", "'\\x%s'::bytea" % ("01" * 16), "S16"),
    ("bytes2", "'\\x%s'::bytea" % ("02" * 16), "S16"),

    ("text1", "'12345'::text", "U5"),
    ("text2", "'12345'::text", "U10"),

    ("dt1", "'1989-01-12 12:00:01.123'::timestamp", "datetime64[s]"),
    ("dt2", "'2022-01-12 12:00:01.123'::timestamp", "datetime64[us]"),
    ("dt3", "'2021-01-12 12:00:01.123'::timestamp", "datetime64[us]"),
    ("dt4", "'2020-01-12 12:00:01.123'::timestamp", "datetime64[us]"),

    ("time1", "'00:01:40'::time", "timedelta64[s]"),
    ("time2", "'01:01:40'::time", "timedelta64[us]"),
]

# lengths = [100, 200, 500, 1000, 1500, 2000]
lengths = [5000, 10000, 15000, 20000, 30000, 50000]


def compose_query(length):
    value_strs = [sql for _, sql, _ in type_samples]
    return f"SELECT {', '.join(value_strs)}\n" \
           f"FROM generate_series(1, {length}) i"


async def fetch_dummy(suite, length):
    if suite.stmt is None:
        suite.stmt = await suite.con.prepare(
            set_query_dtype(compose_query(length),
                            pickle.dumps(np.dtype(np.void))))
    await suite.stmt.fetch()


@pytest.mark.benchmark(
    warmup=True,
    warmup_iterations=2,
)
@pytest.mark.parametrize("length", lengths)
def test_dummy(benchmark, length):
    suite = setup_suite()
    benchmark(_entry, suite, fetch_dummy, length)


async def fetch_record(suite, length):
    if suite.stmt is None:
        suite.stmt = await suite.con.prepare(compose_query(length))
    await suite.stmt.fetch()


@pytest.mark.benchmark(
    warmup=True,
    warmup_iterations=2,
)
@pytest.mark.parametrize("length", lengths)
def test_record(benchmark, length):
    suite = setup_suite()
    benchmark(_entry, suite, fetch_record, length)


async def fetch_numpy_row(suite, length):
    if suite.stmt is None:
        dtype = np.dtype([(name, subdt) for name, _, subdt in type_samples])
        suite.stmt = await suite.con.prepare(
            set_query_dtype(compose_query(length), dtype))
    await suite.stmt.fetch()


@pytest.mark.benchmark(
    warmup=True,
    warmup_iterations=2,
)
@pytest.mark.parametrize("length", lengths)
def test_numpy_row(benchmark, length):
    suite = setup_suite()
    benchmark(_entry, suite, fetch_numpy_row, length)


async def fetch_numpy_column(suite, length):
    if suite.stmt is None:
        dtype = np.dtype([(name, subdt) for name, _, subdt in type_samples],
                         metadata={"blocks": True})
        suite.stmt = await suite.con.prepare(
            set_query_dtype(compose_query(length), dtype))
    await suite.stmt.fetch()


@pytest.mark.benchmark(
    warmup=True,
    warmup_iterations=2,
)
@pytest.mark.parametrize("length", lengths)
def test_numpy_column(benchmark, length):
    suite = setup_suite()
    benchmark(_entry, suite, fetch_numpy_column, length)
